from rank_bm25 import BM25Okapi
import string 
from langchain.text_splitter import RecursiveCharacterTextSplitter
from tqdm.autonotebook import tqdm
import numpy as np
import nltk
nltk.download('stopwords')
from nltk.stem import RSLPStemmer, PorterStemmer
from langchain import OpenAI, LLMChain
import pickle

class DocStore:
    def __init__(self, documents: list):
        self.documents = documents
        self.stop_words = nltk.corpus.stopwords.words()
        self.stemmer = PorterStemmer()
        self.text_splitter = RecursiveCharacterTextSplitter(# Set a really small chunk size, just to show.
            chunk_size = 1000,
            chunk_overlap  = 200,
            length_function = len)
        self.passages = documents
        self.bm25 = BM25Okapi(self.tokenize_corpus())
        with open("./encoded_store.pkl", "wb") as f:
          pickle.dump(self.bm25, f)
        with open("./original_store.pkl", "wb") as f:
          pickle.dump(self.passages, f)

    def chunk_creator(self):
      passages = []
      for document in self.documents:
        texts = self.text_splitter.split_text(document)
        passages.extend(texts) # create flat list of all chunks (loses context across documents)
      return passages
      
    def tokenize_corpus(self):
      tokenize_corpus = []
      for passage in tqdm(self.passages):
        print(passage)
        tokenize_corpus.append(self.bm25_tokenizer(passage))
      return tokenize_corpus

    # We lower case our text and remove stop-words from indexing
    def bm25_tokenizer(self, text):
      tokenized_doc = []
      for token in text.lower().split():
        token = self.stemmer.stem(token) # stem the token
        if len(token) > 0 and token not in self.stop_words: # don't index on stop words 
            tokenized_doc.append(token)
      return tokenized_doc


import asyncio
from langchain import OpenAI, LLMChain, PromptTemplate
import asyncio
from llama_index import Document
from time import time
from typing import Any, List, Optional, Tuple, Union
from langchain.text_splitter import RecursiveCharacterTextSplitter

class SummarizationDocPipeline:
    """A class for doing chunking based on summarization"""
    text_splitter = RecursiveCharacterTextSplitter(# Set a really small chunk size, just to show.
            chunk_size = 750,
            chunk_overlap  = 200,
            length_function = len)
    
    def __init__(self, openai_api_key: str, documents: List[Document]):
      self.openai_api_key = openai_api_key
      self.documents = documents 
      self.chunks, self.references = self.chunk_and_ref_creator()
    
    def chunk_and_ref_creator(self):
      # iterate through the GPT-Index documents 
      texts = []
      references = []
      for doc in self.documents:
        # get the text
        text = doc.text
        # run them through a textsplitter 
        chunks = self.text_splitter.split_text(text)
        # return a list of text + doc_id's 
        texts.extend(chunks)
        references.extend(doc.doc_id*len(chunks))
      return texts, references

    async def async_generate(self, chain, chunk):
      resp = await chain.arun(chunk=chunk)
      print(resp)
      return resp

    async def qa_generate_concurrently(self, openai_api_key, chunks, references, prompt=None):
      import os
      os.environ["OPENAI_API_KEY"] = openai_api_key
      if prompt == None: 
        prompt = "You're an AI assistant to a human, helping them understand the given context. Generate a summary that most appropriately describes this given piece of content: "
      llm = OpenAI(temperature=0.7)
      prompt = PromptTemplate(
          input_variables=["chunk"],
          template= prompt + "{chunk}",
      )
      chain = LLMChain(llm=llm, prompt=prompt)
      results = []
      print(len(chunks))
      if len(chunks) > 100: 
        # batch token limit is 250,000
        chunks = [chunks[i:i+100] for i in range(0, len(chunks), 100)] #chunk in batches of 200 -> each chunk is of size ~1k tokens
        for chunk_idx, chunk in enumerate(chunks): 
          tasks = [self.async_generate(chain, c) for c in chunk]
          result = await asyncio.gather(*tasks)
          results += [Document(c, doc_id = references[idx*chunk_idx])  for idx, c in enumerate(chunk)]
          time.sleep(5)
      else: 
        tasks = [self.async_generate(chain, chunk) for chunk in chunks]
        results = await asyncio.gather(*tasks)
        results = [Document(c, doc_id = references[idx])  for idx, c in enumerate(chunks)]
      return results
    
    def run(self):
      return asyncio.run(self.qa_generate_concurrently(self.openai_api_key, self.chunks, self.references))


from llama_index import GPTSimpleVectorIndex 
from langchain import OpenAI, LLMChain
from langchain.schema import AgentAction, AgentFinish
import re
import os
from langchain.agents import initialize_agent, Tool, ZeroShotAgent, AgentExecutor
from typing import Any, List, Optional, Tuple, Union
from berri_ai.QAAgent import QAAgent

class ComplexInformationQA():
  """Base class for Complex Information QA Agent Class"""  

  def __init__(self, model_key, index = None, prompt = None, functions = [], descriptions = [], return_directly_list = []):
    os.environ["OPENAI_API_KEY"] = model_key
    self.index = index 
    if len(functions) > 0:
      if len(functions) != len(descriptions):
        raise ValueError("The number of functions does not match the number of descriptions!")
      self.functions = functions 
      self.descriptions = descriptions 
      self.return_directly_list = return_directly_list
      tools = self.define_tools()
    else: 
      tools = [
        Tool(
            name = "QueryingDB",
            func=self.querying_db,
            description="This function takes a query string as input and returns the most relevant answer from the documentation as output"
        )]
    
    PREFIX = """Answer the following questions as best you can. You have access to the following tools:"""
    SUFFIX = """Begin!
  
      Question: {input}
      Thought:{agent_scratchpad}"""
    if prompt == None: 
      self.prompt = ZeroShotAgent.create_prompt(
            tools, 
            prefix=PREFIX, 
            suffix=SUFFIX, 
            input_variables=["input", "agent_scratchpad"])
    else: 
      SUFFIX = prompt + "\n" + SUFFIX
      self.prompt = ZeroShotAgent.create_prompt(
        tools, 
        prefix=PREFIX, 
        suffix=SUFFIX, 
        input_variables=["input", "agent_scratchpad"])
    self.tools = tools
    self.llm_chain = LLMChain(llm=OpenAI(temperature=0), prompt=self.prompt)
 
  
  def define_tools(self): 
    agent_tools = []
    for idx, function in enumerate(self.functions): 
      # set tool name
      tool_name = self.extract_function_names(function)
      # set tool func
      tool_func  = function
      # initialize tool for the agent
      agent_tool = Tool(name=tool_name, func=tool_func, description=self.descriptions[idx], return_direct=self.return_directly_list[idx])
      # add to list of tools
      agent_tools.append(agent_tool)
    return agent_tools

  # define function that takes in a list of functions
  def extract_function_names(self, function): 
    # extract the name from the function
    name = function.__name__ 
    # append the name to the list
    return name

  def querying_db(self, query: str):
    response = self.index.query(query)
    response = (response.response, response.source_nodes[0].source_text)
    return response
  
  def run(self, query_string: str):
    llm_chain = LLMChain(llm=OpenAI(temperature=0), prompt=self.prompt)
    agent2 = QAAgent(llm_chain=self.llm_chain, tools=self.tools)
    agent_executor = AgentExecutor.from_agent_and_tools(agent=agent2, tools=self.tools, verbose=True, return_intermediate_steps=True)
    answer = agent_executor({"input":query_string})
    return answer


from langchain import OpenAI, LLMChain
from langchain.schema import AgentAction, AgentFinish
import re
import os
from langchain.agents import initialize_agent, Tool, ZeroShotAgent, AgentExecutor
from typing import Any, List, Optional, Tuple, Union

class QAAgent(ZeroShotAgent):
  """Agent for the MRKL chain."""
  lastDocsViewed: List = None
  FINAL_ANSWER_ACTION = "Final Answer: "
  def _extract_tool_and_input(self, llm_output: str) -> Optional[Tuple[str, str]]:
      llm_output = llm_output.strip()
      if self.FINAL_ANSWER_ACTION in llm_output:
          return "Final Answer", llm_output.split(self.FINAL_ANSWER_ACTION)[-1]

      regex_action = r"Action: (.*?)\n"
      match_action = re.search(regex_action, llm_output)
      action = match_action.group(1)
      if not match_action:
          raise ValueError(f"Could not parse match_action LLM output: `{llm_output}`")
      regex_action_input = r"Action Input: (.*)"
      match_action_input = re.search(regex_action_input, llm_output)
      action_input = match_action_input.group(1)
      if not match_action_input:
          raise ValueError(f"Could not parse match_action_input LLM output: `{llm_output}`")
      return action, action_input.strip(" ").strip('"')

  def doc_extractor(self, observation):
    # assume it's a tuple that can take in 2 inputs -> (the results, the references)
    self.lastDocsViewed = observation[1]
    return observation[0]
    

  def plan(self, intermediate_steps: List[Tuple[AgentAction, str]], **kwargs: Any) -> Union[AgentAction, AgentFinish]:
      """Given input, decided what to do.

      Args:
          intermediate_steps: Steps the LLM has taken to date,
              along with observations
          **kwargs: User inputs.

      Returns:
          Action specifying what tool to use.
      """
      thoughts = ""
      for action, observation in intermediate_steps:
          if type(observation) == tuple:
              observation = self.doc_extractor(observation)
          thoughts += action.log
          thoughts += f"\n{self.observation_prefix}{observation}\n{self.llm_prefix}"
      new_inputs = {"agent_scratchpad": thoughts, "stop": "Observation:"}
      full_inputs = {**kwargs, **new_inputs}
      full_output = self.llm_chain.predict(**full_inputs)
      parsed_output = self._extract_tool_and_input(full_output)
      predict_output = None
      while parsed_output is None:
          full_output = self._fix_text(full_output)
          full_inputs["agent_scratchpad"] += full_output
          output = self.llm_chain.predict(**full_inputs)
          predict_output = output
          full_output += output
          parsed_output = self._extract_tool_and_input(full_output)
      tool, tool_input = parsed_output
      if tool == self.finish_tool_name:
          tool_input = {"response": tool_input, "references": "\n Here is a list of documents that I viewed: " + str(self.lastDocsViewed)}
          return AgentFinish({"output": tool_input}, full_output)
      output = AgentAction(tool, tool_input, full_output)
      return output


PANDAS_CHAIN_DEFAULT_PROMPT = """You are an AI assistant, helping a user formulate pandas expressions based on their queries about the data. The data has been loaded into a pandas dataframe called 'df'. 

Here are your instructions: 
1. For a user's question, explain what you think the user is trying to achieve and write a pandas expression to answer their question. 
2. In cases where a user is asking about a value contained inside a column, you will receive both the column name and the value in the column it is equal to. Do not rewrite it or spell it differently. This will cause your query expression to break
3. When writing the pandas expression, assume the dataframe is called 'df'. Do not call it anything else.
4. Output the value of the pandas expression as a string and do not return the row index (.to_string(index=False))
5. Use Sanity Check to check if the answer is very long (e.g. the entire column). If so, rewrite your expression to only return the head(). 
6. If you are uncertain what the user is asking, say "Hmm, I'm not sure". Do not make things up.

Given these column names: brands_boolean	cat1	cat2	cat3	cat4	cat5	Search Volume	YoY	absolute growth	cluster	cluster_proba	monthly Search Volume	monthly absolute growth	MoM	YoY 1 month absolute growth	YoY 1 month	YoY 24 month absolute growth	YoY 24 month	arrow_YoY	pred_YoY	pred_confidence	pred_YoY_score	cat_ranking	top_affinity_group	top_brands	image	description	filter_cluster	filter_trends_curve	in_out	Search Volume score	Search Volume score_label	pred_YoY score_label	brands Volume	brand count	brands Volume score	brands Volume score_label	concentration score	concentration score_label

Write a pandas expression that answers this query: top cat5 by Search Volume

What do you think the user intent here is? Think step-by-step.

I think the user is trying to find the top category 5 (cat5) items by Search Volume. We can use the `pandas.DataFrame.sort_values()` method to sort the dataframe by the 'Search Volume' column in descending order.

Initial Pandas Expression: df.sort_values('Search Volume', ascending=False)['cat5'].to_string(index=False)

Sanity Check: I believe given user query this expresses the question in pandas correctly. But this will return the entire column, so we might update this expression to only return the head(). 

Final Pandas Expression: df.sort_values('Search Volume', ascending=False)['cat5'].head().to_string(index=False)


Given these column names: brands_boolean	cat1	cat2	cat3	cat4	cat5	Search Volume	YoY	absolute growth	cluster	cluster_proba	monthly Search Volume	monthly absolute growth	MoM	YoY 1 month absolute growth	YoY 1 month	YoY 24 month absolute growth	YoY 24 month	arrow_YoY	pred_YoY	pred_confidence	pred_YoY_score	cat_ranking	top_affinity_group	top_brands	image	description	filter_cluster	filter_trends_curve	in_out	Search Volume score	Search Volume score_label	pred_YoY score_label	brands Volume	brand count	brands Volume score	brands Volume score_label	concentration score	concentration score_label

Write a pandas expression that answers this query: top cat5 by YoY


What do you think the user intent here is? Think step-by-step. It seems like the user is asking for the top values of cat5 by the YoY column.

Pandas Expression: df.sort_values(by='YoY', ascending=False).head(1)['cat5'].to_string(index=False)

Sanity Check: I believe given user query this expresses the question in pandas correctly.

Final Pandas Expression: df.sort_values(by='YoY', ascending=False).head(1)['cat5'].to_string(index=False)


Given these column names: brands_boolean	cat1	cat2	cat3	cat4	cat5	Search Volume	YoY	absolute growth	cluster	cluster_proba	monthly Search Volume	monthly absolute growth	MoM	YoY 1 month absolute growth	YoY 1 month	YoY 24 month absolute growth	YoY 24 month	arrow_YoY	pred_YoY	pred_confidence	pred_YoY_score	cat_ranking	top_affinity_group	top_brands	image	description	filter_cluster	filter_trends_curve	in_out	Search Volume score	Search Volume score_label	pred_YoY score_label	brands Volume	brand count	brands Volume score	brands Volume score_label	concentration score	concentration score_label

Write a pandas expression that answers this query: brands searched alongside hair serum?.When the user says hair serum. It is equal to the 'hair serum' value in the cat5 column.

What do you think the user intent here is? Think step-by-step. I think the user is trying to find the top brands searched alongside hair serum. We can use the `pandas.DataFrame.loc[]` method to filter the dataframe to only rows where the 'cat3' column is equal to 'hair serum'. Then we can use the `pandas.DataFrame.sort_values()` method to sort the dataframe by the 'Search Volume' column in descending order.

Pandas Expression: df.loc[df['cat3'] == 'hair serum'].sort_values('Search Volume', ascending=False)['top_brands'].head().to_string(index=False)

Sanity Check: I believe given the user query, the pandas expression is wrong. It should be looking at cat5 not cat3. 

Final Pandas Expression: df.loc[df['cat5'] == 'hair serum'].sort_values('Search Volume', ascending=False)['top_brands'].head().to_string(index=False)

Given these column names: {sheet_col_names}

Write a pandas expression that answers this query: {user_query}

What do you think the user intent here is? Think step-by-step."""


from rank_bm25 import BM25Okapi
import string 
from langchain.text_splitter import RecursiveCharacterTextSplitter
from tqdm.autonotebook import tqdm
import numpy as np
import nltk
nltk.download('rslp')
nltk.download('stopwords')
from nltk.stem import RSLPStemmer, PorterStemmer
from langchain import OpenAI, LLMChain
import pickle
import os

class DocStore:
    def __init__(self, documents: list, language: str):
        self.documents = documents
        if language == "portugese":
          self.stop_words = nltk.corpus.stopwords.words('portuguese')
          self.stemmer = RSLPStemmer()
        elif language == "english":
          self.stop_words = nltk.corpus.stopwords.words()
          self.stemmer = PorterStemmer()
        self.text_splitter = RecursiveCharacterTextSplitter(# Set a really small chunk size, just to show.
            chunk_size = 1000,
            chunk_overlap  = 200,
            length_function = len)
        self.passages = self.chunk_creator()
        self.bm25 = BM25Okapi(self.tokenize_corpus())
        if not os.path.exists('./berri_files/'):
          os.mkdir("./berri_files/")
        with open("./berri_files/encoded_store.pkl", "wb") as f:
          pickle.dump(self.bm25, f)
        with open("./berri_files/original_store.pkl", "wb") as f:
          pickle.dump(self.passages, f)

    def chunk_creator(self):
      passages = []
      for document in self.documents:
        texts = self.text_splitter.split_text(document)
        passages.extend(texts) # create flat list of all chunks (loses context across documents)
      return passages
      
    def tokenize_corpus(self):
      tokenize_corpus = []
      for passage in tqdm(self.passages):
        tokenize_corpus.append(self.bm25_tokenizer(passage))
      return tokenize_corpus

    # We lower case our text and remove stop-words from indexing
    def bm25_tokenizer(self, text):
      tokenized_doc = []
      for token in text.lower().split():
        token = self.stemmer.stem(token) # stem the token
        if len(token) > 0 and token not in self.stop_words: # don't index on stop words 
            tokenized_doc.append(token)
      return tokenized_doc


from llama_index import PromptHelper, SimpleWebPageReader, GPTSimpleVectorIndex, GPTListIndex, GPTTreeIndex
import os
import requests
import uuid
import json
import base64
from pathlib import Path
from llama_index import download_loader
from llama_index.readers.file.tabular_parser import CSVParser
from langchain.agents import initialize_agent, Tool
from langchain import OpenAI
from llama_index import Document
from langchain import PromptTemplate
import re
from langchain.chains import LLMChain
import asyncio
import pandas as pd
import fuzzywuzzy
from fuzzywuzzy import process
import numpy as np
from berri_ai.agents.DataAnalysisPrompt import PANDAS_CHAIN_DEFAULT_PROMPT
from berri_ai.search_strategies.bm25.DocStore_CSV import DocStore
from berri_ai.ComplexInformationQA import ComplexInformationQA
import os

os.environ[
  "OPENAI_API_KEY"] = "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"


class DataAnalysisQA:

  def __init__(self,
               df,
               openai_api_key,
               additional_functions=None,
               additional_descriptions=None):
    self.df = df
    self.openai_api_key = openai_api_key
    if additional_functions != None and additional_descriptions != None and len(
        additional_functions) == len(additional_descriptions):
      self.additional_functions = additional_functions
      self.additional_descriptions = additional_descriptions
    self.unique_values = []
    for column in self.df.columns:
      unique_val = self.df[column].unique().ravel()
      self.unique_values.extend(unique_val)

  def likely_col(self, keywords):
    output = ""
    if len(keywords) > 1:
      output += ". These are the keywords from the user's query: " + ' '.join(
        keywords)
    unique_values = []
    for column in self.df.columns:
      unique_val = self.df[column].unique().ravel()
      unique_values.extend(unique_val)
    no_col = True
    for keyword in keywords:
      potential_val_cols = []
      scores = process.extract(keyword, unique_values)
      scores = set(x for x in scores if x[1] > 90)
      if len(scores) > 0:
        for score in scores:
          # get all the potential columns this value could belong to
          for col in self.df.columns:
            if self.df[col].isin([score[0]]).any():
              potential_val_cols.append((score[0], col))
      print(potential_val_cols)
      if len(potential_val_cols) > 0:
        output_str = ". When the user says " + keyword + ' , they are most likely referring to one of these values / inside these columns: '
        print(scores)
        for idx, potential_val_col in enumerate(potential_val_cols):
          output_str += '. Value ' + str(idx) + ': ' + str(
            potential_val_col[0]) + ' Column ' + str(idx) + ': ' + str(
              potential_val_col[1])
        output += output_str
        no_col = False
    if no_col:
      output = ""
    return output

  def keyword_extractor(self, query_str):
    llm = OpenAI(temperature=0.7)
    prefix = """A user is asking questions about a spreadsheet. Extract the keywords from the user query that might make sense as names of columns or values in a spreadsheet. If you are not certain, say 'I am not certain', do not make things up. If there are multiple keywords, separate them with a comma (,).
    \n
    User Query: Top cat5 by YoY
    Keywords: cat5, YoY 
    \n
    User Query: How many sku's do we have for variant handles?
    Keywords: sku, variant handle 
    \n
    User Query: Top cat5 by YoY that is very low competition
    Keywords: cat5, YoY, very low competition
    \n
    User Query: """
    input = prefix + query_str
    input += "\n Keywords: "
    return llm(input)

  def check_if_col_or_not_v_2(self, user_query):
    # use llm to get keywords
    keywords = self.keyword_extractor(user_query)
    keywords = keywords.split(",")
    updated_query = user_query
    # use fuzzy match to compare keyword in keywords to column names and see if there's a likely match
    values = []
    for keyword in keywords:
      # List of tuples containing the similarity score and the string
      scores = process.extract(keyword, self.df.columns.unique().ravel())

      # Sort the list of tuples by similarity score
      sorted_scores = sorted(scores, key=lambda x: x[1], reverse=True)
      if int(
          sorted_scores[0][1]
      ) < 85:  # if the keyword match is below 85/100, then it's probably a value inside a column, and not a column name.
        values.append(keyword.strip())
      elif int(sorted_scores[0][1]) >= 90:
        updated_query += '. When the user says ' + keyword + ' they are most likely referring to this column: ' + sorted_scores[
          0][0] + '.'
      print("sorted_scores: ", sorted_scores)
    return updated_query, values

  def sanity_checker(self, user_intent, pandas_expression):
    prompt_template = """
      You are an AI assistant, helping a user formulate pandas expressions based on their queries about the data. Ensure the given pandas expression is a valid pandas expression and aligns with the user's intent. If 'Initial Pandas Expression' aligns to 'User Intent' and is a valid pandas expression, simply output the same in 'Final Pandas Expression', do not change anything.
    
      Format: 
      User Intent: The question the user is trying to ask about their data. 
      Initial Pandas Expression: An initial approach to answering this question, expressed in pandas. 
      Thought: Think if the initial pandas expression expresses the user's intent correctly.
      Final Pandas Expression: If you think the initial pandas expression expression correctly, output the same as Initial Pandas Expression. Else re-write expression to better express the user's intent.
    
      User Intent: {user_intent}
    
      Initial Pandas Expression: {initial_pandas_expression}
    
      Thought: 
      """
    prompt = PromptTemplate(
      input_variables=["user_intent", "initial_pandas_expression"],
      template=prompt_template)
    formatted_prompt = prompt.format(
      user_intent=user_intent, initial_pandas_expression=pandas_expression)
    llm = OpenAI(temperature=0.7, best_of=3)
    pd_output = llm(formatted_prompt)
    return pd_output

  def pandas_chain(self, user_query, sheet_col_names):
    prompt_template = """
      You are an AI assistant, helping a user formulate pandas expressions based on their queries about the data. The data has been loaded into a pandas dataframe called 'df'. 
    
    Here are your instructions: 
    1. For a user's question, explain what you think the user is trying to achieve and write a pandas expression to answer their question. 
    2. In cases where a user is asking about a value contained inside a column, you will receive both the column name and the value in the column it is equal to. Do not rewrite it or spell it differently. This will cause your query expression to break
    3. When writing the pandas expression, assume the dataframe is called 'df'. Do not call it anything else.
    4. Output the value of the pandas expression as a string and do not return the row index (.to_string(index=False))
    5. Use Sanity Check to check if the answer is very long (e.g. the entire column). If so, rewrite your expression to only return the head(). 
    6. If you are uncertain what the user is asking, say "Hmm, I'm not sure". Do not make things up.
    
    Given these column names: brands_boolean	cat1	cat2	cat3	cat4	cat5	Search Volume	YoY	absolute growth	cluster	cluster_proba	monthly Search Volume	monthly absolute growth	MoM	YoY 1 month absolute growth	YoY 1 month	YoY 24 month absolute growth	YoY 24 month	arrow_YoY	pred_YoY	pred_confidence	pred_YoY_score	cat_ranking	top_affinity_group	top_brands	image	description	filter_cluster	filter_trends_curve	in_out	Search Volume score	Search Volume score_label	pred_YoY score_label	brands Volume	brand count	brands Volume score	brands Volume score_label	concentration score	concentration score_label
    
    Write a pandas expression that answers this query: top cat5 by Search Volume
    
    What do you think the user intent here is? Think step-by-step.
    
    I think the user is trying to find the top category 5 (cat5) items by Search Volume. We can use the `pandas.DataFrame.sort_values()` method to sort the dataframe by the 'Search Volume' column in descending order.
    
    Initial Pandas Expression: df.sort_values('Search Volume', ascending=False)['cat5'].to_string(index=False)
    
    
    Given these column names: brands_boolean	cat1	cat2	cat3	cat4	cat5	Search Volume	YoY	absolute growth	cluster	cluster_proba	monthly Search Volume	monthly absolute growth	MoM	YoY 1 month absolute growth	YoY 1 month	YoY 24 month absolute growth	YoY 24 month	arrow_YoY	pred_YoY	pred_confidence	pred_YoY_score	cat_ranking	top_affinity_group	top_brands	image	description	filter_cluster	filter_trends_curve	in_out	Search Volume score	Search Volume score_label	pred_YoY score_label	brands Volume	brand count	brands Volume score	brands Volume score_label	concentration score	concentration score_label
    
    Write a pandas expression that answers this query: top cat5 by YoY
    
    
    What do you think the user intent here is? Think step-by-step. It seems like the user is asking for the top values of cat5 by the YoY column.
    
    Initial Pandas Expression: df.sort_values(by='YoY', ascending=False).head(1)['cat5'].to_string(index=False)
    
    
    Given these column names: brands_boolean	cat1	cat2	cat3	cat4	cat5	Search Volume	YoY	absolute growth	cluster	cluster_proba	monthly Search Volume	monthly absolute growth	MoM	YoY 1 month absolute growth	YoY 1 month	YoY 24 month absolute growth	YoY 24 month	arrow_YoY	pred_YoY	pred_confidence	pred_YoY_score	cat_ranking	top_affinity_group	top_brands	image	description	filter_cluster	filter_trends_curve	in_out	Search Volume score	Search Volume score_label	pred_YoY score_label	brands Volume	brand count	brands Volume score	brands Volume score_label	concentration score	concentration score_label
    
    Write a pandas expression that answers this query: brands searched alongside hair serum?.When the user says hair serum. It is equal to the 'hair serum' value in the cat5 column.
    
    What do you think the user intent here is? Think step-by-step. I think the user is trying to find the top brands searched alongside hair serum. We can use the `pandas.DataFrame.loc[]` method to filter the dataframe to only rows where the 'cat3' column is equal to 'hair serum'. Then we can use the `pandas.DataFrame.sort_values()` method to sort the dataframe by the 'Search Volume' column in descending order.
    
    Initial Pandas Expression: df.loc[df['cat3'] == 'hair serum'].sort_values('Search Volume', ascending=False)['top_brands'].head().to_string(index=False)
    
    Given these column names: {sheet_col_names}
    
    Write a pandas expression that answers this query: {user_query}
    
    What do you think the user intent here is? Think step-by-step.
      """
    prompt = PromptTemplate(input_variables=["sheet_col_names", "user_query"],
                            template=prompt_template)

    formatted_prompt = prompt.format(sheet_col_names=sheet_col_names,
                                     user_query=user_query)
    llm = OpenAI(temperature=0.7, best_of=3)
    pd_output = llm(formatted_prompt)
    return pd_output

  def conditional_search(self, query):
    sheet_col_names = ", ".join([col for col in self.df.columns])
    pd_output = self.pandas_chain(user_query=query,
                                  sheet_col_names=sheet_col_names)
    # extract pandas expression
    if "Initial Pandas Expression" in pd_output:
      # run eval
      print(pd_output)
      pandas_expression = pd_output.split(
        "Initial Pandas Expression:")[1].strip()
      final_pd_output = self.sanity_checker(query, pandas_expression)
      print(final_pd_output)
      pandas_expression = final_pd_output.split(
        "Final Pandas Expression:")[1].strip()
      match = re.search("(?:`)?df(.*)", pandas_expression).group(1)
      if match:
        # match = match.replace("df", "df")
        df_expression = "self.df" + match
        print("df_expression: ", df_expression)
        value = eval(df_expression)
        return (value,
                pd_output + "\n Post Sanity Checker: " + final_pd_output)
      else:
        return pd_output
    else:
      return pd_output

  def query(self, user_input: str):

    # check if column or not
    user_input, values = self.check_if_col_or_not_v_2(user_input)
    print(values)
    likely_columns = ""
    if len(values) > 0:
      likely_columns = self.likely_col(values)
      print("likely columns: ", likely_columns)
    # pass in query + context into conditional search and get answer
    final_input_str = user_input + "." + likely_columns
    print("final_input_str: ", final_input_str)
    output = self.conditional_search(final_input_str)
    if len(output) == 2:
      response = output[0]
      references = "I think the keywords mentioned were: " + ",".join(values) + ". And my final input into the pandas expression model was: " + final_input_str + ". And this is what the model came up with: " + output[1]
      return {"response": response, "references": references}
    return {"response": "I am not certain, could you rephrase that?", "references": "I checked if your question had keywords that either belonged to a column or could possibly be a value in the spreadsheet"}


# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['berri_ai',
 'berri_ai.agents',
 'berri_ai.search_strategies.bm25',
 'berri_ai.search_strategies.pipelines']

package_data = \
{'': ['*']}

install_requires = \
['fuzzywuzzy>=0.18.0,<0.19.0',
 'html2text>=2020.1.16,<2021.0.0',
 'llama-index==0.4.24',
 'mixpanel>=4.10.0,<5.0.0',
 'nltk>=3.8.1,<4.0.0',
 'numpy>=1.18,<2.0',
 'openai>=0.26.5,<0.27.0',
 'pipreqs>=0.4.11,<0.5.0',
 'rank-bm25>=0.2.2,<0.3.0',
 'requests>=2.28.2,<3.0.0',
 'tiktoken>=0.2.0,<0.3.0',
 'tqdm>=4.64.1,<5.0.0']

setup_kwargs = {
    'name': 'berri-ai',
    'version': '0.20.9',
    'description': '',
    'long_description': None,
    'author': 'Team Berri',
    'author_email': 'clerkieai@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8.1,<4.0.0',
}


setup(**setup_kwargs)



__version__ = '0.20.9'
import ast 
import re
import os 
import traceback
import time
import pipreqs
import subprocess
import requests
import shutil
from mixpanel import Mixpanel
import uuid
from typing import Union, List, Optional
import json 

mp = Mixpanel("69c4620538e80373ae4ef8edb32ce5e3")

def send_files(user_email, template_name):
    url = f'https://berriserverv2.krrishdholakia.repl.co/berri_orchestrator?user_email={user_email}&template_name={template_name}'
    files_list = []
    
    for file in os.listdir("./berri_files"):
        if os.path.isdir(file):
          raise ValueError(f"🚨🚨 Deployment Error 📣: {file} is a folder, it can't be sent. Join us on Discord (https://discord.gg/KvG3azf39U) if this is causing you issues.")
        with open(os.path.join("./berri_files", file), 'rb') as f:
          files_list.append((file, f.read()))
            
    response = requests.get(url, files=files_list)
    return response.json()["message"]

def traverse_ast(tree_dict, code, tree, global_dict):
  for node in tree.body:
      if isinstance(node, ast.For):
          node_value_source = ast.get_source_segment(code, node)
          for child in node.body: 
            if isinstance(child, ast.Assign):
              value_source = ast.get_source_segment(code, child)
              str_split = value_source.split("=", 1)
              key_val = str(str_split[0]).strip()
              if key_val in tree_dict: 
                if node_value_source == tree_dict[key_val]:
                  continue
                else:
                  raise KeyError("🚨🚨 Deployment Error 📣: The variable - " + key_val + ' has been initialized 2 times. Here are the instances: \n 1.' + tree_dict[key_val][:100] + '\n 2.' + value_source[:100])
              tree_dict[key_val] = node_value_source
      if isinstance(node, ast.With):
        node_value_source = ast.get_source_segment(code, node)
        for child in node.body: 
          if isinstance(child, ast.Assign):
            value_source = ast.get_source_segment(code, child)
            str_split = value_source.split("=", 1)
            key_val = str(str_split[0]).strip()
            key_val = key_val.split(",") # if there's multiple items being assigned on the same line
            for kv in key_val:
              kv = kv.strip()
              if kv in tree_dict: 
                if node_value_source == tree_dict[kv]:
                  continue
                else:
                  raise KeyError("🚨🚨 Deployment Error 📣: The variable - " + kv + ' has been initialized 2 times. Here are the instances: \n 1.' + tree_dict[kv][:100] + '\n 2.' + value_source[:100])
              tree_dict[kv] = node_value_source
      if isinstance(node, ast.Assign):
        value_source = ast.get_source_segment(code, node)
        str_split = value_source.split("=", 1)
        key_val = str(str_split[0]).strip()
        key_val = key_val.split(",") # if there's multiple items being assigned on the same line
        for kv in key_val:
          kv = kv.strip()
          if kv in tree_dict: 
            if value_source == tree_dict[kv]:
              continue
            else:
              raise KeyError("🚨🚨 Deployment Error 📣: The variable - " + kv + ' has been initialized 2 times. Here are the instances: \n 1.' + tree_dict[kv][:100] + '\n 2.' + value_source[:100])
          tree_dict[kv] = value_source
      elif isinstance(node, ast.ClassDef):
          key_val = node.name.strip()
          if key_val in tree_dict: 
            if ast.get_source_segment(code, node) == tree_dict[key_val]:
              continue
            else:
              raise KeyError("🚨🚨 Deployment Error 📣: The variable - " + key_val + ' has been initialized 2 times. Here are the instances: \n 1.' + tree_dict[key_val][:100] + '\n 2.' + value_source[:100])
          tree_dict[key_val] = ast.get_source_segment(code, node)
      elif isinstance(node, ast.FunctionDef):
          key_val = node.name.strip()
          if key_val in tree_dict: 
            if ast.get_source_segment(code, node) == tree_dict[key_val]:
              continue
            else:
              raise KeyError("🚨🚨 Deployment Error 📣: The variable - " + key_val + ' has been initialized 2 times. Here are the instances: \n 1.' + tree_dict[key_val][:100] + '\n 2.' + value_source[:100])
          tree_dict[key_val] = ast.get_source_segment(code, node)
      elif isinstance(node, ast.Import):
          value_source = ast.get_source_segment(code, node)
          for name in node.names:
            key_val = name.name
            tree_dict[name.name] = value_source
            if "." in name.name:
              split_names = name.name.split(".")[0]
              key_val = split_names
              tree_dict[key_val] = value_source
      elif isinstance(node, ast.ImportFrom):
          value_source = ast.get_source_segment(code, node)
          for name in node.names:
            key_val = name.name
            tree_dict[name.name] = value_source

def check_requirements(package):
    with open('./berri_files/requirements.txt', 'r') as file:
        content = file.read()
        if package not in content:
            with open('./berri_files/requirements.txt', 'a') as file2:
                file2.write(package)

def get_dependencies(code_segment_list, global_dict):
    # Run function in a sandbox and catch ImportErrors
    dep_modules = None
    bool_val = True
    i = 0
    while bool_val:
      if i >= 20:
        raise Exception("there is an issue in the loop")
      dep_modules = []
      try:
        for code_segment in code_segment_list:
          print("code_segment being executed: ", code_segment)
          exec(code_segment, global_dict)
        bool_val = False
      except ModuleNotFoundError as e:
        print(e)
        traceback.print_exc()
        text = e.args[0]
        match = re.search(r"'(.*)'", text)
        if match:
            package_name = match.group(1)
            check_requirements(package_name)
            install_statements = []
            install_statements.append("import subprocess")
            install_statements.append(str("""subprocess.call(['pip', 'install','""" + package_name + "'])"))
            for install_statement in install_statements:
              exec(install_statement)
      except Exception as e:
        print(e)
        traceback.print_exc()
        if hasattr(e, 'name'):
          dep_modules.append(e.name)  # Add module that caused error
          bool_val = False
        else:
            text = e.args[0]
            pattern = r"pip install (\w+)"
            if isinstance(text, list):
              text = str(text[0])
            match = re.search(pattern, text)
            if match:
              package_name = match.group(1)
              install_statements = []
              install_statements.append("import subprocess")
              install_statements.append(str("""subprocess.call(['pip', 'install','""" + package_name + "'])"))
              for install_statement in install_statements:
                exec(install_statement)
            else:
              match = re.search(r"'(.*)'", text)
              if match:
                  quoted_text = match.group(1)
                  dep_modules.append(quoted_text)
                  bool_val = False
      print("bool_val: ", bool_val)
    return dep_modules

def copy_files():
  current_directory = os.getcwd()
  if not os.path.exists(current_directory + '/berri_files'):
      os.mkdir(current_directory + '/berri_files')
  list_of_restricted_dirs = ['berri_files', 'sample_data', 'drive', '.config', '__pycache__', '.ipynb_checkpoints', '.modules']
  for subdir, dirs, files in os.walk(current_directory):
    for file in files:
      if (file != 'sample_data' and file != 'drive') and (subdir == current_directory):
        # print("file: ", file)
        shutil.copy(file, current_directory + '/berri_files')
    for dir in dirs:
      if (dir not in list_of_restricted_dirs) and (subdir == current_directory):
        # print("dir: ", dir)
        shutil.copytree(dir, current_directory + '/berri_files/' + dir)

def run_loop(environment_list, code_list, tree_dict, global_dict):
  code_segment_list = None
  for i in range(20):
    code_segment_list = environment_list + code_list 
    dependencies = get_dependencies(code_segment_list, global_dict)
    if len(dependencies) == 0:
      break
    for dependency in dependencies:
      code = tree_dict[dependency]
      code_list.insert(0, code)
  return code_segment_list


def save_requirements(path, requirements):
  subprocess.run(["pipreqs", "--mode", "no-pin", path])
  with open('./berri_files/requirements.txt', 'r') as f:
    existing_reqs = f.readlines()
    existing_reqs = [req.strip().split("=")[0] for req in existing_reqs]

  with open('./berri_files/requirements.txt', 'a') as f:
    if len(requirements) > 0:
      for requirement in requirements:
        for r in requirement:
          tmp_req = r.split("=")[0]
          if tmp_req not in existing_reqs:
              f.write(tmp_req + '\n')
          else:
              continue

def get_requirements(line):
  line_split = line.split(' && ')
  requirements = []

  for line in line_split:
    line_split_2 = line.split(' ')
    package = line_split_2[-1]
    requirements.append(package)
  
  return requirements

def install_requirements():
  with open('./berri_files/requirements.txt', 'r') as f:
    existing_reqs = f.readlines()
    existing_reqs = [req.strip().split("=")[0] for req in existing_reqs]

  for package_name in existing_reqs:
    install_statements = []
    install_statements.append("import subprocess")
    install_statements.append(str("""subprocess.call(['pip', 'install','""" + package_name + "'])"))
    for install_statement in install_statements:
      exec(install_statement)


def deploy_func(user_email: str, executing_function, test_str: str):
  from google.colab import _message
  try:
    print("Begun deployment..")
    print("🚨 Hit an error? let us know in the Discord (https://discord.gg/KvG3azf39U).")
    print("🐍 Converting notebook to python and generating requirements.txt, this might take 1-2 minutes.")

    try:
      mp.track(str(uuid.uuid4()), "package.start.berri.deploy_func()", {
        'UserEmail': user_email
      })
    except:
      print("MP error")
    # assume you're in a google colab 
    if not os.path.exists('./berri_files/'):
      os.mkdir("./berri_files/")

    copy_files() # copies all local non-drive/sample_data files and folders into berri_files

    # Obtain the notebook JSON as a string
    notebook_json_string = _message.blocking_request('get_ipynb', request='', timeout_sec=5)

    # save to temporary file
    lines = []
    requirements = []
    for cell in notebook_json_string["ipynb"]["cells"]:
      if cell["cell_type"] == "code":
        for line in cell["source"]:
          if line.startswith("!pip install"):
            requirements.append(get_requirements(line))
            # continue
          elif not line.startswith("!"):
            lines.append(line)
    
    f = open("./berri_files/agent_code.py", "w")
    for line in lines:
      # line = read_file_copy_drive_files(line)
      f.write(line + "\n")
    f.close()
    
    # save the requirements to a requirements.txt file 
    save_requirements("./berri_files", requirements)

    with open("./berri_files/agent_code.py") as f:
        lines = f.readlines()
    
    with open("./berri_files/agent_code.py") as f:
        code = f.read()

    tree = ast.parse(code)
    tree_dict = {}
    global_dict = {}
    # traverse the file and create a dictionary
    traverse_ast(tree_dict, code, tree, global_dict)
    # print("global_dict: ", global_dict)
    print("tree_dict: ", tree_dict)


    # set the executing line
    agent_executing_line = "" + str(executing_function) + "('" + test_str+ "')"
    # print("agent_executing_line: ", agent_executing_line)
    # find the import statements
    import_statements = []
    for line in lines:
      if "google.colab" in line or "drive.mount('/content/drive')" in line:
        continue
      elif "import" in line:
        import_statements.append(line)

    # run_loop(import_statements, tree_dict, global_dict)
    # print(import_statements)

    parent_dependencies = []
    
    for key in tree_dict:
      if "os" in key:
        parent_dependencies.append(tree_dict[key])
    
    # print("parent_dependencies: ", parent_dependencies)
    
    parent_dependencies = run_loop(import_statements, parent_dependencies, tree_dict, global_dict)

    # print("updated parent dependencies: ", parent_dependencies)
        
    implementation_code = [agent_executing_line]
    environment_setup_list = import_statements + parent_dependencies
    # print("running main code now")
    install_requirements()
    all_up_code = run_loop(environment_setup_list, implementation_code, tree_dict, global_dict)

    # print("all_up_code: ", all_up_code)
    # print("executing_function: ", executing_function)
    with open("./berri_files/agent_code.py", "w") as f:
      for code_segment in all_up_code:
        if executing_function in code_segment:
          # print("true")
          code_segment = code_segment.replace(executing_function, "agent").strip()
        f.write(code_segment + "\n")

    print("😱 Building docker image.. this might take 1-2 minutes")
    endpoint = "https://" + send_files(user_email, "berri_backend_server_template")

    print("🚧 Currently deploying to [NOT READY YET] 👉 " + endpoint)
    print("⌛️ It'll be ready in 15 mins. We'll email you  @ " + user_email)
  except Exception as e:
    print(e)
    traceback.print_exc()
    print("🚨🚨 Deployment Error 📣: There was an error deploying your project. Join us on Discord (https://discord.gg/KvG3azf39U) and we'll fix this for you.")
    try:
      mp.track(str(uuid.uuid4()), "package.error.berri.wrapper_func_deploy()", {
        'UserEmail': user_email,
        'Error': e
      })
    except:
      print("MP error")
  print("=====================")
  print("Got feedback? Text/WhatsApp us 👉 +17708783106")

def deploy_gpt_index(user_email: str):
  from google.colab import _message
  try:
    print("Begun deployment..")
    print("🚨 Hit an error? let us know in the Discord (https://discord.gg/KvG3azf39U).")
    print("🐍 Converting notebook to python and generating requirements.txt, this might take 1-2 minutes.")

    try:
      mp.track(str(uuid.uuid4()), "package.start.berri.gpt_index_deploy()", {
        'UserEmail': user_email
      })
    except:
      print("MP error")
    # assume you're in a google colab 
    if not os.path.exists('./berri_files/'):
      os.mkdir("./berri_files/")

    copy_files() # copies all local non-drive/sample_data files and folders into berri_files

    # Obtain the notebook JSON as a string
    notebook_json_string = _message.blocking_request('get_ipynb', request='', timeout_sec=5)

    # save to temporary file
    lines = []
    requirements = []
    for cell in notebook_json_string["ipynb"]["cells"]:
      if cell["cell_type"] == "code":
        for line in cell["source"]:
          if line.startswith("!pip install"):
            requirements.append(get_requirements(line))
            # continue
          elif not line.startswith("!"):
            lines.append(line)
    
    f = open("./berri_files/agent_code.py", "w")
    for line in lines:
      if ".query(" in line:
        # print("line where query was found: ", line)
        # get the variable name
        print("line: ", line)
        if "=" in line:
          initialization_line = line.split("=", 1)[1]
        else:
          initialization_line = line
        print("initialization_line: ", initialization_line)
        # print("original_agent_name: ", original_agent_name)
        # replace with new name 
        prev_name = initialization_line.split(".query")[0].strip()
        print("prev_name: ", prev_name)
        reinitialize_agent = "agent = " + prev_name
        f.write(reinitialize_agent + "\n")
        line = initialization_line.replace(prev_name, "agent").strip()
        print("new line: ", line)
        # print(initialization_line)
        f.write(line + "\n")
      else:
        # line = read_file_copy_drive_files(line)
        f.write(line + "\n")
    f.close()
    
    # save the requirements to a requirements.txt file 
    save_requirements("./berri_files", requirements)

    with open("./berri_files/agent_code.py") as f:
        lines = f.readlines()
    
    with open("./berri_files/agent_code.py") as f:
        code = f.read()

    tree = ast.parse(code)
    tree_dict = {}
    global_dict = {}
    # traverse the file and create a dictionary
    traverse_ast(tree_dict, code, tree, global_dict)
    # print("global_dict: ", global_dict)
    # print("tree_dict: ", tree_dict)

    # find the executing line
    agent_executing_line = "agent.query("
    # find the import statements
    import_statements = []
    for line in lines:
      if agent_executing_line in line:
        # print(line)
        agent_executing_line = line.strip() # find the executing line 
        # print("agent_executing_line: ", agent_executing_line)
      elif "google.colab" in line or "drive.mount('/content/drive')" in line:
        continue
      elif "import" in line:
        import_statements.append(line)

    # run_loop(import_statements, tree_dict, global_dict)
    # print(import_statements)

    parent_dependencies = []
    
    for key in tree_dict:
      if "os" in key:
        parent_dependencies.append(tree_dict[key])
    
    # print("parent_dependencies: ", parent_dependencies)
    
    parent_dependencies = run_loop(import_statements, parent_dependencies, tree_dict, global_dict)

    # print("updated parent dependencies: ", parent_dependencies)
        
    implementation_code = [agent_executing_line]
    environment_setup_list = import_statements + parent_dependencies
    # print("running main code now")
    all_up_code = run_loop(environment_setup_list, implementation_code, tree_dict, global_dict)

    # print("all_up_code: ", all_up_code)

    with open("./berri_files/agent_code.py", "w") as f:
      for code_segment in all_up_code:
        f.write(code_segment + "\n")

    print("😱 Building docker image.. this might take 1-2 minutes")
    endpoint = "https://" + send_files(user_email, "berri_backend_server_template")

    print("🚧 Currently deploying to [NOT READY YET] 👉 " + endpoint)
    print("⌛️ It'll be ready in 15 mins. We'll email you  @ " + user_email)
  except Exception as e:
    print(e)
    traceback.print_exc()
    print("🚨🚨 Deployment Error 📣: There was an error deploying your project. Join us on Discord (https://discord.gg/KvG3azf39U) and we'll fix this for you.")
    try:
      mp.track(str(uuid.uuid4()), "package.error.berri.gpt_index_deploy()", {
        'UserEmail': user_email,
        'Error': e
      })
    except:
      print("MP error")
  print("=====================")
  print("Got feedback? Text/WhatsApp us 👉 +17708783106")

def deploy(user_email: str):
  from google.colab import _message
  try:
    print("Begun deployment..")
    print("🚨 Hit an error? let us know in the Discord (https://discord.gg/KvG3azf39U).")
    print("🐍 Converting notebook to python and generating requirements.txt, this might take 1-2 minutes.")
    try:
      mp.track(str(uuid.uuid4()), "package.start.berri.deploy()", {
        'UserEmail': user_email
      })
    except:
      print("MP error")

    # assume you're in a google colab 
    if not os.path.exists('./berri_files/'):
      os.mkdir("./berri_files/")

    copy_files() # copies all local non-drive/sample_data files and folders into berri_files

    # Obtain the notebook JSON as a string
    notebook_json_string = _message.blocking_request('get_ipynb', request='', timeout_sec=5)

    # save to temporary file
    lines = []
    requirements = []
    for cell in notebook_json_string["ipynb"]["cells"]:
      if cell["cell_type"] == "code":
        for line in cell["source"]:
          if line.startswith("!pip install"):
            requirements.append(get_requirements(line))
            # continue
          elif not line.startswith("!"):
            lines.append(line)
    
  # get the variable name
  # replace with new name 
    original_agent_name = None
    f = open("./berri_files/agent_code.py", "w")
    for line in lines:
      if original_agent_name and original_agent_name + ".run(" in line: # if it exists and is being executed, replace that line
        line = line.replace(original_agent_name + ".run(", "agent.run(")
        f.write(line + "\n")
      if "initialize_agent(" in line:
        initialization_line = line.split("=", 1)
        original_agent_name = initialization_line[0].strip() # get the original agent variable name
        # print("original_agent_name: ", original_agent_name)
        initialization_line = "agent = " + initialization_line[1]
        # print(initialization_line)
        f.write(initialization_line + "\n")
      else:
        # line = read_file_copy_drive_files(line)
        f.write(line + "\n")
    f.close()
    
    # save the requirements to a requirements.txt file 
    save_requirements("./berri_files", requirements)

    # if len(requirements) > 0: 
    #   with open('./berri_files/requirements.txt', 'a') as f:
    #     for requirement in requirements:
    #       for r in requirement:
    #         print("requirement: ", r)
    #         f.write("\n" + r + '\n')

    with open("./berri_files/agent_code.py") as f:
        lines = f.readlines()
    
    with open("./berri_files/agent_code.py") as f:
        code = f.read()

    tree = ast.parse(code)
    tree_dict = {}
    global_dict = {}
    # traverse the file and create a dictionary
    traverse_ast(tree_dict, code, tree, global_dict)
    # print("global_dict: ", global_dict)
    # print("tree_dict: ", tree_dict)

    # find the executing line
    agent_executing_line = "agent.run("
    # find the import statements
    import_statements = []
    for line in lines:
      if agent_executing_line in line:
        # print(line)
        agent_executing_line = line.strip() # find the executing line 
      elif "google.colab" in line or "drive.mount('/content/drive')" in line:
        continue
      elif "import" in line:
        import_statements.append(line)

    # run_loop(import_statements, tree_dict, global_dict)
    # print(import_statements)

    parent_dependencies = []
    
    for key in tree_dict:
      if "os" in key:
        parent_dependencies.append(tree_dict[key])
    
    # print("parent_dependencies: ", parent_dependencies)
    
    parent_dependencies = run_loop(import_statements, parent_dependencies, tree_dict, global_dict)

    # print("updated parent dependencies: ", parent_dependencies)
        
    implementation_code = [agent_executing_line]
    environment_setup_list = import_statements + parent_dependencies
    # print("running main code now")
    all_up_code = run_loop(environment_setup_list, implementation_code, tree_dict, global_dict)

    # print("all_up_code: ", all_up_code)

    with open("./berri_files/agent_code.py", "w") as f:
      for code_segment in all_up_code:
        f.write(code_segment + "\n")

    print("😱 Building docker image.. this might take 1-2 minutes")
    endpoint = "https://" + send_files(user_email, "berri_backend_server_template")

    print("🚧 Currently deploying to [NOT READY YET] 👉 " + endpoint)
    print("⌛️ It'll be ready in 15 mins. We'll email you  @ " + user_email)
  except Exception as e:
    print(e)
    traceback.print_exc()
    print("🚨🚨 Deployment Error 📣: There was an error deploying your project. Join us on Discord (https://discord.gg/KvG3azf39U) and we'll fix this for you.")
    try:
      mp.track(str(uuid.uuid4()), "package.error.berri.deploy()", {
        'UserEmail': user_email,
        'Error': e
      })
    except:
      print("MP error")
  print("=====================")
  print("Got feedback? Text/WhatsApp us 👉 +17708783106")